var ometa = require('ometajs'),
    OMeta = ometa.OMeta,
    Parser = ometa.Parser,
    BSJSParser = exports.BSJSParser = ometa.BSJSParser,
    BSJSTranslator = exports.BSJSTranslator = ometa.BSJSTranslator,
    crypto = require('crypto'),
    sha1 = function(o) { return crypto.createHash('sha1').update(o).digest('base64') };

function Identifier () {
    this.counter = 0;
    this.cache = {};
}

Identifier.prototype.identify = function(o) {
    var key = JSON.stringify(o);
    return key in this.cache ?
        this.cache[key] :
        this.cache[key] = ++this.counter;
};

exports.compile = function(templatesAndOther) {

    var templates = templatesAndOther[1];

    function newPredicMemo(o, k, v) {
        var r = {};
        for(var i in o) r[i] = o[i];
        r[k] = v;
        return r;
    }

    var predicatesValues = (function() {
            var vals = {};

            templates.forEach(function(t){
                t[0].forEach(function(subMatch){
                    var p = subMatch[0],
                        c = JSON.stringify(subMatch[2]);
                    (vals[p] || (vals[p] = {}))[c] = subMatch[2];
                });
            });

            for(var p in vals) {
                var vs = [];
                for(var v in vals[p]) vs.push(vals[p][v]);
                vals[p] = vs;
            }

            return vals;
        })();

    var cache = {},
        idCounter = 1;

    function merge(o) {
        var hash,
            id = idCounter++;

        if(o['switch']) {
            hash += 'switch ' + JSON.stringify(o['switch']) + '{';
            o.cases.forEach(function(c){
                c[1] = merge(c[1]);
                hash += c[1].hash + " ";
            });
            o['default'] = merge(o['default']);
            hash += o['default'].hash + "}";
            hash = sha1(hash);
        } else {
            hash = o.hash || sha1(JSON.stringify(o.exprs || o.stmt)));
        }

        (hash in cache) || (cache[o.hash = hash] = o);

        cache[hash].id = id;

        return cache[hash];
    }

    function doTemplate(i, j, predicMemo) {
        var template = templates[i];
        if(!template) return merge({ 'stmt': ['throw', ['get', 'true']] });

        var subMatch = template[0][j];
        if(!subMatch) return {
            //comment: JSON.stringify([i, j, predicMemo]),
            stmt: template[1] };

        var jj = j, s;
        while(s = template[0][++jj]) {
            var predicate = s[0],
                predicateConst = s[2];

            if(predicate in predicMemo)
                if(predicMemo[predicate] != JSON.stringify(predicateConst))
                    return doTemplate(i + 1, 0, predicMemo);
        }

        var predicate = subMatch[0],
            predicateConst = subMatch[2];

        if(predicate in predicMemo)
            if(predicMemo[predicate] == JSON.stringify(predicateConst))
                return doTemplate(i, j + 1, predicMemo);
            else
                return doTemplate(i + 1, 0, predicMemo);
        else
            return merge({
                //'comment': JSON.stringify([i, j, predicMemo]),
                'switch': subMatch[1],
                'cases': predicatesValues[predicate].map(function(v){
                        return [v, doTemplate(i, j, newPredicMemo(predicMemo, predicate, JSON.stringify(v)))]
                    }),
                'default': doTemplate(i, j, newPredicMemo(predicMemo, predicate, undefined))
            });
    }


    var merges = {};

    function serialize(o, tails) {
        if(o.id in merges) {
            tails[o.id] = o;
            return 'break l' + o.id + ';';
        }

        var res = '';

        o.comment && (res += '/*' + o.comment + '*/\n');
        o.id && (res += '/*' + o.id + '*/');

        if(o['switch']) {
            if(o.cases.length == 1) {
                var c = o.cases[0];
                res += 'if(' +
                    XJSTCompiler.match(o['switch'], 'trans') +
                    ' === ' +
                    XJSTCompiler.match(c[0], 'trans') + ') {\n' +
                        serialize(c[1], tails) +
                    '} else {\n' +
                        serialize(o['default'], tails) +
                    '}';
            } else {
                res += 'switch(' + XJSTCompiler.match(o['switch'], 'trans') + ') {\n' +
                    o.cases.map(function(c){
                        return 'case ' + XJSTCompiler.match(c[0], 'trans') + ':\n' +
                            serialize(c[1], tails) + '; break;'
                        }).join('\n') +
                    'default: ' + serialize(o['default'], tails) +
                    '}\n';
            }
        } else {
            res += o.exprs ?
                'return ' + XJSTCompiler.match(o.exprs, 'tBody') :
                XJSTCompiler.match(o.stmt, 'trans') + ';return;';
        }
        return res;
    }

    function shiftTails(tails) {
        var ids = [];
        for(var i in tails) ids.push(i);
        ids.sort(function(a, b) { return a - b });
        var res = tails[ids[0]];
        delete tails[ids[0]];
        return res;
    }

    function detectJoins(o, joins) {
        if('id' in o) {
            if(o.id in joins) {
                joins[o.id]++;
            } else {
                joins[o.id] = 1;
                if(o['switch']) {
                    o.cases.forEach(function(c){ detectJoins(c[1], joins) });
                    detectJoins(o['default'], joins);
                }
            }
        }
        return joins;
    }

    function serializeTop(prog) {
        var res = 'exports.apply = function(c){\n',
            joins = detectJoins(prog, {});

        var labels = [];
        for(var k in joins) {
            if(joins[k] == 1) continue
            labels.push(k);
            merges[k] = true;
        }
        labels.sort(function(a, b) { return b - a });

        res += labels.map(function(l){ return 'l' + l + ': {\n' }).join('');

        var tails = {};
        tails[prog.id] = prog;
        var t, first = true;
        while(t = shiftTails(tails)) {
            first || (res += '}');
            first = false;
            delete merges[t.id];
            res += serialize(t, tails);
        }

        return res += '};';
    }

    return '(function(exports){' +
        XJSTCompiler.match(templatesAndOther[0], 'other') + ';' +
        serializeTop(doTemplate(0, 0, {})) +
        'return exports})(typeof exports === "undefined"? {} : exports)';

};
